/*
 * Copyright 2014-present MongoDB, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <php.h>

#include "phongo.h"
#include "Unserializable_arginfo.h"

zend_class_entry* phongo_unserializable_ce;

static int phongo_implement_unserializable(zend_class_entry* interface, zend_class_entry* class_type)
{
	if (class_type->ce_flags & ZEND_ACC_ENUM) {
		zend_class_entry* error_interface = interface;

		/* Persistable extends Unserializable, so this handler may be called for
		 * Persistable enums; prefer Persistable in the fatal message when applicable. */
		if (interface == phongo_unserializable_ce && phongo_persistable_ce && instanceof_function(class_type, phongo_persistable_ce)) {
			error_interface = phongo_persistable_ce;
		}

		zend_error_noreturn(E_ERROR, "Enum class %s cannot implement interface %s", ZSTR_VAL(class_type->name), ZSTR_VAL(error_interface->name));
		return FAILURE;
	}

	return SUCCESS;
}

void phongo_unserializable_init_ce(INIT_FUNC_ARGS)
{
	phongo_unserializable_ce                             = register_class_MongoDB_BSON_Unserializable();
	phongo_unserializable_ce->interface_gets_implemented = phongo_implement_unserializable;
}
